// Copyright (c) 2018 Team Xecuter
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.
#include <string.h>
#include "usbfs.h"

static Service g_usbFsSrv;
static u64 g_refCnt;

Result usbFsIsInitialized() {
    if (serviceIsActive(&g_usbFsSrv)) {
        return 0;
    }

    return 0x666;
}

Result usbFsInitialize(void) {
    Result rc;
    atomicIncrement64(&g_refCnt);

    if (serviceIsActive(&g_usbFsSrv)) {
        return 0;
    }

#ifndef HAVE_USBFS
    rc = smUnregisterService("usbfs");
    if (rc == 0xE15) {
        // Service not exist
        return rc;
    }
#endif

    rc = smGetService(&g_usbFsSrv, "usbfs");
    if (R_FAILED(rc)) {
        return rc;
    }

    return rc;
}

void usbFsExit(void) {
    if (atomicDecrement64(&g_refCnt) == 0) {
        serviceClose(&g_usbFsSrv);
    }
}


Result usbFsIsReady() {
    IpcCommand c;
    ipcInitialize(&c);

    struct {
        u64 magic;
        u64 cmd_id;
    } *raw;

    raw = ipcPrepareHeader(&c, sizeof(*raw));

    raw->magic = SFCI_MAGIC;
    raw->cmd_id = UsbFs_Cmd_IsReady;

    Result rc = serviceIpcDispatch(&g_usbFsSrv);

    if (R_SUCCEEDED(rc)) {
        IpcParsedCommand r;
        ipcParse(&r);

        struct {
            u64 magic;
            u64 result;
        } *resp = r.Raw;

        rc = resp->result;
    }

    return rc;
}


Result usbFsGetMountStatus(u64* status) {
    IpcCommand c;
    ipcInitialize(&c);

    struct {
        u64 magic;
        u64 cmd_id;
    } *raw;

    raw = ipcPrepareHeader(&c, sizeof(*raw));

    raw->magic = SFCI_MAGIC;
    raw->cmd_id = UsbFs_Cmd_GetMountStatus;

    Result rc = serviceIpcDispatch(&g_usbFsSrv);

    if (R_SUCCEEDED(rc)) {
        IpcParsedCommand r;
        ipcParse(&r);

        struct {
            u64 magic;
            u64 result;
            u64 status;
        } *resp = r.Raw;

        rc = resp->result;
        if (R_SUCCEEDED(rc)) {
            *status = resp->status;
        }
    }

    return rc;
}


Result usbFsOpenFile(u64* fileid, const char* filepath, u64 mode) {
    IpcCommand c;
    ipcInitialize(&c);
    ipcAddSendBuffer(&c, filepath, strlen(filepath) + 1, BufferType_Normal);

    struct {
        u64 magic;
        u64 cmd_id;
        u64 mode;
    } *raw;

    raw = ipcPrepareHeader(&c, sizeof(*raw));

    raw->magic = SFCI_MAGIC;
    raw->cmd_id = UsbFs_Cmd_OpenFile;
    raw->mode = mode;

    Result rc = serviceIpcDispatch(&g_usbFsSrv);

    if (R_SUCCEEDED(rc)) {
        IpcParsedCommand r;
        ipcParse(&r);

        struct {
            u64 magic;
            u64 result;
            u64 fileid;
        } *resp = r.Raw;

        rc = resp->result;
        if (R_SUCCEEDED(rc)) {
            *fileid = resp->fileid;
        }
    }

    return rc;
}



Result usbFsCloseFile(u64 fileid) {
    IpcCommand c;
    ipcInitialize(&c);

    struct {
        u64 magic;
        u64 cmd_id;
        u64 fileid;
    } *raw;

    raw = ipcPrepareHeader(&c, sizeof(*raw));

    raw->magic = SFCI_MAGIC;
    raw->cmd_id = UsbFs_Cmd_CloseFile;
    raw->fileid = fileid;

    Result rc = serviceIpcDispatch(&g_usbFsSrv);

    if (R_SUCCEEDED(rc)) {
        IpcParsedCommand r;
        ipcParse(&r);

        struct {
            u64 magic;
            u64 result;
        } *resp = r.Raw;

        rc = resp->result;
    }

    return rc;
}


Result usbFsReadFile(u64 fileid, void* buffer, size_t size, size_t* retsize) {
    IpcCommand c;
    ipcInitialize(&c);
    ipcAddRecvBuffer(&c, buffer, size, BufferType_Normal);

    struct {
        u64 magic;
        u64 cmd_id;
        u64 fileid;
    } *raw;

    raw = ipcPrepareHeader(&c, sizeof(*raw));

    raw->magic = SFCI_MAGIC;
    raw->cmd_id = UsbFs_Cmd_ReadFile;
    raw->fileid = fileid;

    Result rc = serviceIpcDispatch(&g_usbFsSrv);

    if (R_SUCCEEDED(rc)) {
        IpcParsedCommand r;
        ipcParse(&r);

        struct {
            u64 magic;
            u64 result;
            u64 retsize;
        } *resp = r.Raw;

        rc = resp->result;
        if (R_SUCCEEDED(rc)) {
            *retsize = resp->retsize;
        }
    }

    return rc;
}

Result usbFsWriteFile(u64 fileid, const void* buffer, size_t size, size_t* retsize) {
    IpcCommand c;
    ipcInitialize(&c);
    ipcAddSendBuffer(&c, buffer, size, BufferType_Normal);

    struct {
        u64 magic;
        u64 cmd_id;
        u64 fileid;
    } *raw;

    raw = ipcPrepareHeader(&c, sizeof(*raw));

    raw->magic = SFCI_MAGIC;
    raw->cmd_id = UsbFs_Cmd_WriteFile;
    raw->fileid = fileid;

    Result rc = serviceIpcDispatch(&g_usbFsSrv);

    if (R_SUCCEEDED(rc)) {
        IpcParsedCommand r;
        ipcParse(&r);

        struct {
            u64 magic;
            u64 result;
            u64 retsize;
        } *resp = r.Raw;

        rc = resp->result;
        if (R_SUCCEEDED(rc)) {
            *retsize = resp->retsize;
        }
    }

    return rc;
}

Result usbFsSeekFile(u64 fileid, u64 pos, u64 whence, u64 *retpos) {
    IpcCommand c;
    ipcInitialize(&c);

    struct {
        u64 magic;
        u64 cmd_id;
        u64 fileid;
        u64 pos;
        u64 whence;
    } *raw;

    raw = ipcPrepareHeader(&c, sizeof(*raw));

    raw->magic = SFCI_MAGIC;
    raw->cmd_id = UsbFs_Cmd_SeekFile;
    raw->fileid = fileid;
    raw->pos = pos;
    raw->whence = whence;

    Result rc = serviceIpcDispatch(&g_usbFsSrv);

    if (R_SUCCEEDED(rc)) {
        IpcParsedCommand r;
        ipcParse(&r);

        struct {
            u64 magic;
            u64 result;
            u64 retpos;
        } *resp = r.Raw;

        rc = resp->result;
        if (R_SUCCEEDED(rc)) {
            *retpos = resp->retpos;
        }
    }

    return rc;
}

Result usbFsSyncFile(u64 fileid) {
    IpcCommand c;
    ipcInitialize(&c);

    struct {
        u64 magic;
        u64 cmd_id;
        u64 fileid;
    } *raw;

    raw = ipcPrepareHeader(&c, sizeof(*raw));

    raw->magic = SFCI_MAGIC;
    raw->cmd_id = UsbFs_Cmd_SyncFile;
    raw->fileid = fileid;

    Result rc = serviceIpcDispatch(&g_usbFsSrv);

    if (R_SUCCEEDED(rc)) {
        IpcParsedCommand r;
        ipcParse(&r);

        struct {
            u64 magic;
            u64 result;
        } *resp = r.Raw;

        rc = resp->result;
    }

    return rc;
}


Result usbFsTruncateFile(u64 fileid, u64 size) {
    IpcCommand c;
    ipcInitialize(&c);

    struct {
        u64 magic;
        u64 cmd_id;
        u64 fileid;
        u64 size;
    } *raw;

    raw = ipcPrepareHeader(&c, sizeof(*raw));

    raw->magic = SFCI_MAGIC;
    raw->cmd_id = UsbFs_Cmd_TruncateFile;
    raw->fileid = fileid;
    raw->size = size;

    Result rc = serviceIpcDispatch(&g_usbFsSrv);

    if (R_SUCCEEDED(rc)) {
        IpcParsedCommand r;
        ipcParse(&r);

        struct {
            u64 magic;
            u64 result;
        } *resp = r.Raw;

        rc = resp->result;
    }

    return rc;
}

Result usbFsDeleteFile(const char* filepath) {
    IpcCommand c;
    ipcInitialize(&c);
    ipcAddSendBuffer(&c, filepath, strlen(filepath) + 1, BufferType_Normal);

    struct {
        u64 magic;
        u64 cmd_id;
    } *raw;

    raw = ipcPrepareHeader(&c, sizeof(*raw));

    raw->magic = SFCI_MAGIC;
    raw->cmd_id = UsbFs_Cmd_DeleteFile;

    Result rc = serviceIpcDispatch(&g_usbFsSrv);

    if (R_SUCCEEDED(rc)) {
        IpcParsedCommand r;
        ipcParse(&r);

        struct {
            u64 magic;
            u64 result;
        } *resp = r.Raw;

        rc = resp->result;
    }

    return rc;
}


Result usbFsStatFile(u64 fileid, u64* size, u64* mode) {
    IpcCommand c;
    ipcInitialize(&c);

    struct {
        u64 magic;
        u64 cmd_id;
        u64 fileid;
    } *raw;

    raw = ipcPrepareHeader(&c, sizeof(*raw));

    raw->magic = SFCI_MAGIC;
    raw->cmd_id = UsbFs_Cmd_StatFile;
    raw->fileid = fileid;

    Result rc = serviceIpcDispatch(&g_usbFsSrv);

    if (R_SUCCEEDED(rc)) {
        IpcParsedCommand r;
        ipcParse(&r);

        struct {
            u64 magic;
            u64 result;
            u64 size;
            u64 mode;
        } *resp = r.Raw;

        rc = resp->result;
        if (R_SUCCEEDED(rc)) {
            *size = resp->size;
            *mode = resp->mode;
        }
    }

    return rc;
}

Result usbFsStatPath(const char* path, u64* size, u64* mode) {
    IpcCommand c;
    ipcInitialize(&c);
    ipcAddSendBuffer(&c, path, strlen(path) + 1, BufferType_Normal);

    struct {
        u64 magic;
        u64 cmd_id;
    } *raw;

    raw = ipcPrepareHeader(&c, sizeof(*raw));

    raw->magic = SFCI_MAGIC;
    raw->cmd_id = UsbFs_Cmd_StatPath;

    Result rc = serviceIpcDispatch(&g_usbFsSrv);

    if (R_SUCCEEDED(rc)) {
        IpcParsedCommand r;
        ipcParse(&r);

        struct {
            u64 magic;
            u64 result;
            u64 size;
            u64 mode;
        } *resp = r.Raw;

        rc = resp->result;
        if (R_SUCCEEDED(rc)) {
            *size = resp->size;
            *mode = resp->mode;
        }
    }

    return rc;
}


Result usbFsStatFilesystem(u64* totalsize, u64* freesize) {
    IpcCommand c;
    ipcInitialize(&c);

    struct {
        u64 magic;
        u64 cmd_id;
    } *raw;

    raw = ipcPrepareHeader(&c, sizeof(*raw));

    raw->magic = SFCI_MAGIC;
    raw->cmd_id = UsbFs_Cmd_StatFilesystem;

    Result rc = serviceIpcDispatch(&g_usbFsSrv);

    if (R_SUCCEEDED(rc)) {
        IpcParsedCommand r;
        ipcParse(&r);

        struct {
            u64 magic;
            u64 result;
            u64 totalsize;
            u64 freesize;
        } *resp = r.Raw;

        rc = resp->result;
        if (R_SUCCEEDED(rc)) {
            *totalsize = resp->totalsize;
            *freesize = resp->freesize;
        }
    }

    return rc;
}

Result usbFsOpenDir(u64* dirid, const char* dirpath) {
    IpcCommand c;
    ipcInitialize(&c);
    ipcAddSendBuffer(&c, dirpath, strlen(dirpath) + 1, BufferType_Normal);

    struct {
        u64 magic;
        u64 cmd_id;
    } *raw;

    raw = ipcPrepareHeader(&c, sizeof(*raw));

    raw->magic = SFCI_MAGIC;
    raw->cmd_id = UsbFs_Cmd_OpenDir;

    Result rc = serviceIpcDispatch(&g_usbFsSrv);

    if (R_SUCCEEDED(rc)) {
        IpcParsedCommand r;
        ipcParse(&r);

        struct {
            u64 magic;
            u64 result;
            u64 dirid;
        } *resp = r.Raw;

        rc = resp->result;
        if (R_SUCCEEDED(rc)) {
            *dirid = resp->dirid;
        }
    }

    return rc;
}

Result usbFsCloseDir(u64 dirid) {
    IpcCommand c;
    ipcInitialize(&c);

    struct {
        u64 magic;
        u64 cmd_id;
        u64 dirid;
    } *raw;

    raw = ipcPrepareHeader(&c, sizeof(*raw));

    raw->magic = SFCI_MAGIC;
    raw->cmd_id = UsbFs_Cmd_CloseDir;
    raw->dirid = dirid;

    Result rc = serviceIpcDispatch(&g_usbFsSrv);

    if (R_SUCCEEDED(rc)) {
        IpcParsedCommand r;
        ipcParse(&r);

        struct {
            u64 magic;
            u64 result;
        } *resp = r.Raw;

        rc = resp->result;
    }

    return rc;
}


Result usbFsReadDir(u64 dirid, u64* type, u64* size, char* name, size_t namemax) {
    IpcCommand c;
    ipcInitialize(&c);
    ipcAddRecvBuffer(&c, name, namemax, BufferType_Normal);

    struct {
        u64 magic;
        u64 cmd_id;
        u64 dirid;
    } *raw;

    raw = ipcPrepareHeader(&c, sizeof(*raw));

    raw->magic = SFCI_MAGIC;
    raw->cmd_id = UsbFs_Cmd_ReadDir;
    raw->dirid = dirid;

    Result rc = serviceIpcDispatch(&g_usbFsSrv);

    if (R_SUCCEEDED(rc)) {
        IpcParsedCommand r;
        ipcParse(&r);

        struct {
            u64 magic;
            u64 result;
            u64 type;
            u64 size;
        } *resp = r.Raw;

        rc = resp->result;
        if (R_SUCCEEDED(rc)) {
            *type = resp->type;
            *size = resp->size;
        }
    }

    return rc;
}


Result usbFsCreateDir(const char* dirpath) {
    IpcCommand c;
    ipcInitialize(&c);
    ipcAddSendBuffer(&c, dirpath, strlen(dirpath) + 1, BufferType_Normal);

    struct {
        u64 magic;
        u64 cmd_id;
    } *raw;

    raw = ipcPrepareHeader(&c, sizeof(*raw));

    raw->magic = SFCI_MAGIC;
    raw->cmd_id = UsbFs_Cmd_CreateDir;

    Result rc = serviceIpcDispatch(&g_usbFsSrv);

    if (R_SUCCEEDED(rc)) {
        IpcParsedCommand r;
        ipcParse(&r);

        struct {
            u64 magic;
            u64 result;
        } *resp = r.Raw;

        rc = resp->result;
    }

    return rc;
}

Result usbFsDeleteDir(const char* dirpath) {
    IpcCommand c;
    ipcInitialize(&c);
    ipcAddSendBuffer(&c, dirpath, strlen(dirpath) + 1, BufferType_Normal);

    struct {
        u64 magic;
        u64 cmd_id;
    } *raw;

    raw = ipcPrepareHeader(&c, sizeof(*raw));

    raw->magic = SFCI_MAGIC;
    raw->cmd_id = UsbFs_Cmd_DeleteDir;

    Result rc = serviceIpcDispatch(&g_usbFsSrv);

    if (R_SUCCEEDED(rc)) {
        IpcParsedCommand r;
        ipcParse(&r);

        struct {
            u64 magic;
            u64 result;
        } *resp = r.Raw;

        rc = resp->result;
    }

    return rc;
}

Result usbFsReadRaw(u64 sector, u64 sectorcount, void* buffer) {
    IpcCommand c;
    ipcInitialize(&c);
    ipcAddRecvBuffer(&c, buffer, 0x200ULL * sectorcount, BufferType_Type1);

    struct {
        u64 magic;
        u64 cmd_id;
        u64 sector;
        u64 sectorcount;
    } *raw;

    raw = ipcPrepareHeader(&c, sizeof(*raw));

    raw->magic = SFCI_MAGIC;
    raw->cmd_id = UsbFs_Cmd_ReadRaw;
    raw->sector = sector;
    raw->sectorcount = sectorcount;

    Result rc = serviceIpcDispatch(&g_usbFsSrv);

    if (R_SUCCEEDED(rc)) {
        IpcParsedCommand r;
        ipcParse(&r);

        struct {
            u64 magic;
            u64 result;
        } *resp = r.Raw;

        rc = resp->result;
    }

    return rc;
}
